#include <Arduino.h>
#include <AccelStepper.h>
#include <ESP8266WiFi.h>
#include <PubSubClient.h>
#include <configuration.h>

AccelStepper stepper[numSteppers];
int stepperStatus[numSteppers];

int targetPosition[numSteppers];
int targetTransparency[numSteppers];

uint32 stepperCalibrationCounter[numSteppers];

WiFiClient espClient;
PubSubClient client(espClient);

void logText(const char* message);
void connectWifi();
void connectMqtt();
void mqttCallback(char* topic, byte* payload, unsigned int length);
void setPositionAndTransparency(int position, int stepper, int transparency);
void setPosition(int stepper, int position);
void setTransparency(int stepper, int transparency);
int getPosition(int stepper);
int getTransparency(int stepper);

void setup() {
    Serial.begin(9600);

    //Stepper Setup
    for(int i = 0; i<numSteppers;i++){
        stepper[i] = AccelStepper(AccelStepper::FULL2WIRE, stepperPinDefinitions[i][1], stepperPinDefinitions[i][2]);
        pinMode(stepperPinDefinitions[i][0], OUTPUT);
        digitalWrite(stepperPinDefinitions[i][0], HIGH);
        stepper[i].setMaxSpeed(5000);
        stepper[i].setAcceleration(2000);
        stepperStatus[i]=0;
    }

    Serial.println("[WIFI] Starting...");
    Serial.print("[WIFI] Connecting to ");
    Serial.println(WIFI_SSID);

    WiFi.mode(WIFI_STA);
    WiFi.begin(WIFI_SSID, WIFI_PASS);

    connectWifi();

    client.setServer(MQTT_BROKER, MQTT_PORT);
    client.setCallback(mqttCallback);

    connectMqtt();

}


void loop() {
    if(WiFi.status() != WL_CONNECTED){
        connectWifi();
    }

    if (!client.connected()) {
        connectMqtt();
    }

    client.loop();

    for(int i = 0; i < numSteppers; i++){
        //Homing Sequence
        switch(stepperStatus[i]){
            case 0:{
                digitalWrite(stepperPinDefinitions[i][0], LOW);
                stepper[i].move(-curtainClosedSteps[i][numCurtainClosedPositions/2]);
                stepperStatus[i]=1;
                break;
            }
            case 1:{
                if(!stepper[i].isRunning()){
                    stepperStatus[i]=2;
                }
                break;
            }
            case 2: {
                digitalWrite(stepperPinDefinitions[i][0], HIGH);
                stepperCalibrationCounter[i] = millis();
                stepperStatus[i] = 3;
                break;
            }
            case 3: {
                if(millis()>stepperCalibrationCounter[i]+5000){
                    stepper[i].setCurrentPosition(0);
                    client.publish(("stat/"+String(MQTT_NAME)+"/"+String(i)+"/position").c_str(), String(0).c_str());
                    client.publish(("stat/"+String(MQTT_NAME)+"/"+String(i)+"/transparency").c_str(), String(0).c_str());
                    stepperStatus[i] = 4;
                }
                break;
            }
        }
        if(!stepper[i].isRunning() && (targetPosition[i] == 100 || targetPosition[i] == 0) && targetTransparency[i] == 0){
            digitalWrite(stepperPinDefinitions[i][0], HIGH);
        }else{
            digitalWrite(stepperPinDefinitions[i][0], LOW);
        }
        stepper[i].run();
    }
}

void logText(const char* message){
    Serial.println(message);
    client.publish(("tele/"+String(MQTT_NAME)).c_str(), message);
}

void connectWifi(){
    Serial.print("[WIFI] Connecting");
    while(WiFi.status() != WL_CONNECTED){
        delay(500);
        Serial.print(".");
    }
    Serial.println("");
    Serial.print("[WIFI] Connected to WiFi. IP: ");
    Serial.println(WiFi.localIP());
}

void connectMqtt(){
    Serial.print("[MQTT] Connecting...");
    while (!client.connected()) {
        if (!client.connect(MQTT_NAME, MQTT_USER, MQTT_PASS)) {
            Serial.print("(failed, rc=");
            Serial.print(client.state());
            Serial.print(")");
            delay(5000);
        }
        
    }
    Serial.println("");
    Serial.println("[MQTT] Connected");
    for(int i = 0; i < numSteppers; i++){
        client.subscribe(("cmnd/"+String(MQTT_NAME)+"/"+String(i)+"/position").c_str());
        client.subscribe(("cmnd/"+String(MQTT_NAME)+"/"+String(i)+"/transparency").c_str());
        client.subscribe(("cmnd/"+String(MQTT_NAME)+"/"+String(i)+"/command").c_str());
        client.subscribe(("cmnd/"+String(MQTT_NAME)+"/"+String(i)+"/steps").c_str());
    }
}

void mqttCallback(char* topic, byte* payload, unsigned int length){
    String msg = "";
    for (int i = 0; i < length; i++) {
        msg += (char)payload[i];
    }

    Serial.print("[MQTT] Received message on ");
    Serial.print(topic);
    Serial.print(": ");
    Serial.println(msg);

    for(int i = 0; i < numSteppers; i++){
        if(String(topic).equals(("cmnd/"+String(MQTT_NAME)+"/"+String(i)+"/command").c_str())){
            if(msg.equals("OPEN")){
                setPositionAndTransparency(i, 100, 0);
            }else if(msg.equals("CLOSE")){
                setPositionAndTransparency(i, 0, 0);
            }else if(msg.equals("STOP")){
                stepper[i].moveTo(stepper[i].currentPosition());
            }
        }else if(String(topic).equals(("cmnd/"+String(MQTT_NAME)+"/"+String(i)+"/position").c_str())){
            setPosition(i, msg.toInt());
        }else if(String(topic).equals(("cmnd/"+String(MQTT_NAME)+"/"+String(i)+"/transparency").c_str())){
            setTransparency(i, msg.toInt());
        }
    }
}

void updatePositionAndTransparency(int stepperId){
    if(stepperStatus[stepperId] != 4){return;}
    int position = targetPosition[stepperId];
    int transparency = targetTransparency[stepperId];
    int closedPosition = map(position, 0, 100, 0, numCurtainClosedPositions-1);
    int closedPositionSteps = curtainClosedSteps[stepperId][closedPosition];
    int nextPositionSteps = position!=100?curtainClosedSteps[stepperId][closedPosition+1]:closedPositionSteps;
    int transparencyOffsetSteps = map(transparency, 0, 100, 0, abs(closedPositionSteps-nextPositionSteps)/2);
    int targetStep = curtainClosedSteps[stepperId][closedPosition]+transparencyOffsetSteps;
    stepper[stepperId].moveTo(targetStep);
}
void setPositionAndTransparency(int stepperId, int position, int transparency){
    targetPosition[stepperId]=position;
    targetTransparency[stepperId]=transparency;
    updatePositionAndTransparency(stepperId);
    client.publish(("stat/"+String(MQTT_NAME)+"/"+String(stepperId)+"/position").c_str(), String(position).c_str());
    client.publish(("stat/"+String(MQTT_NAME)+"/"+String(stepperId)+"/transparency").c_str(), String(transparency).c_str());
}
void setPosition(int stepperId, int position){
    targetPosition[stepperId]=position;
    updatePositionAndTransparency(stepperId);
    client.publish(("stat/"+String(MQTT_NAME)+"/"+String(stepperId)+"/position").c_str(), String(position).c_str());
}
void setTransparency(int stepperId, int transparency){
    targetTransparency[stepperId]=transparency;
    updatePositionAndTransparency(stepperId);
    client.publish(("stat/"+String(MQTT_NAME)+"/"+String(stepperId)+"/transparency").c_str(), String(transparency).c_str());
}

int getPosition(int stepperId){
    return targetPosition[stepperId];
}
int getTransparency(int stepperId){
    return targetTransparency[stepperId];
}